/*
 * arch/arm/plat-cxd41xx/cachectrl.c
 *
 * Cache lockdown driver
 *
 * Copyright 2010,2011 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/smp.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <asm/uaccess.h>

#define PROC_NAME "cachectrl"
#define WAY_MASK 0xf

void __attribute__((weak)) l2cachectrl_write(char *arg, size_t len)
{
	printk("ERROR:L2 cache is not supported.\n");
}

int __attribute__((weak)) l2cachectrl_show(struct seq_file *seq, void *v)
{
	return 0;
}


static DEFINE_SPINLOCK(cachectrl_lock);

static void lockdown_dcache(int mask)
{
	unsigned long flags;

	mask &= WAY_MASK;
	local_irq_save(flags);
	asm("mrc	p15, 0, r0, c9, c0, 0\n"
	    "bic	r0, r0, #0xf\n"
	    "orr	r0, r0, %0\n"
	    "mov	r1, #0\n"
	    "mcr	p15, 0, r1, c7, c14, 0\n"
	    "mcr	p15, 0, r0, c9, c0, 0\n"
	    :: "r" (mask):"cc");
	local_irq_restore(flags);
}

/* ASSUME: cpu affinity is set.
 *   Ex.)  taskset 4 sh -c "echo 7 > /proc/cachectrl"
 */
static ssize_t cachectrl_write(struct file *file, const char __user *buf,
			       size_t len, loff_t *ppos)
{
	char arg[64];
	int n, val;
	extern void lockdown_dcache(int);

	if (len > sizeof arg - 1) {
		printk(KERN_ERR "/proc/%s: too long\n", PROC_NAME);
		return -1;
	}
	if (copy_from_user(arg, buf, len)) {
		return -EFAULT;
	}
	arg[len] = '\0';

	if (!strncmp(arg, "L2 ", 3)) {
		l2cachectrl_write(arg + 3, len - 3);
		return len;
	}
	n = sscanf(arg, "%x", &val);
	if (n != 1){
		printk(KERN_ERR "/proc/%s: invalid format\n", PROC_NAME);
		return -1;
	}
	if (val < 0  ||  val > 0xf){
		printk(KERN_ERR "/proc/%s: data out of range\n", PROC_NAME);
		return -1;
	}

	lockdown_dcache(val);

	return len;
}

#define CACHECTRL_NR_SEQ 2

static void *cachectrl_seq_start(struct seq_file *seq, loff_t *pos)
{
	if (*pos < CACHECTRL_NR_SEQ)
		return pos;
	return NULL;
}

static void *cachectrl_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	(*pos)++;
	if (*pos < CACHECTRL_NR_SEQ)
		return pos;
	return NULL;
}

static void cachectrl_seq_stop(struct seq_file *seq, void *v)
{
}

/* ASSUME: set cpu affinity is already done.
 *   Ex.)  taskset 4 "cat /proc/cachectrl"
 */
static int cachectrl_seq_show(struct seq_file *seq, void *v)
{
	int i = *(loff_t *)v;
	int cpu;
	uint dat, aux;

	if (i > 0) {
		return l2cachectrl_show(seq, v);
	}
	spin_lock(&cachectrl_lock);
	cpu = smp_processor_id();
	asm("mrc	p15, 0, %0, c9, c0, 0" :"=r" (dat)::"cc");
	asm("mrc	p15, 0, %0, c1, c0, 1" :"=r" (aux)::"cc");
	spin_unlock(&cachectrl_lock);
	dat &= WAY_MASK;
	seq_printf(seq, "CPU%d: D-Cache lockdown: 0x%x\n", cpu, dat);
	seq_printf(seq, "CPU%d: AUX reg         : 0x%x\n", cpu, aux);
	return 0;
}

static struct seq_operations cachectrl_seq_ops = {
	.start	= cachectrl_seq_start,
	.next	= cachectrl_seq_next,
	.stop	= cachectrl_seq_stop,
	.show	= cachectrl_seq_show,
};

static int cachectrl_seq_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &cachectrl_seq_ops);
}

static struct file_operations proc_cachectrl_fops = {
	.owner		= THIS_MODULE,
	.open		= cachectrl_seq_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
	.write		= cachectrl_write,
};

static struct proc_dir_entry *proc_cachectrl = NULL;

static int __init cachectrl_init(void)
{
	proc_cachectrl = create_proc_entry(PROC_NAME, 0, NULL);
	if (!proc_cachectrl) {
		printk(KERN_ERR "cachectrl: can not create proc entry\n");
		return -1;
	}
	proc_cachectrl->proc_fops = &proc_cachectrl_fops;
	return 0;
}

module_init(cachectrl_init);
